#!/usr/bin/env python
#
# @package      hubzero-submit-monitors
# @file         monitorJob.py
# @author       Steven Clark <clarks@purdue.edu>
# @copyright    Copyright (c) 2004-2014 HUBzero Foundation, LLC.
# @license      http://www.gnu.org/licenses/lgpl-3.0.html LGPLv3
#
# Copyright (c) 2004-2014 HUBzero Foundation, LLC.
#
# This file is part of: The HUBzero(R) Platform for Scientific Collaboration
#
# The HUBzero(R) Platform for Scientific Collaboration (HUBzero) is free
# software: you can redistribute it and/or modify it under the terms of
# the GNU Lesser General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# HUBzero is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# HUBzero is a registered trademark of HUBzero Foundation, LLC.
#
# ----------------------------------------------------------------------
#  monitorJob.py
#
#  script which gathers monitoring statistics from remote sites and
#  reports job status to distributor on demand
#
import sys
import os
import subprocess
import shlex
import select
import signal
import socket
import time
import logging
from errno import EINTR

from hubzero.submit.LogMessage          import getLogPIDMessage as getLogMessage
from hubzero.submit.MonitorsInfo        import MonitorsInfo
from hubzero.submit.JobInfo             import JobInfo
from hubzero.submit.JobMonitor          import JobMonitor
from hubzero.submit.RemoteTunnelMonitor import RemoteTunnelMonitor

JOBMONITORHOST    = ""
JOBMONITORPORT    = 5727
TUNNELMONITORHOST = ""
TUNNELMONITORPORT = 5729

MONITORROOT = os.path.join(os.sep,'opt','submit')

MONITORLOGLOCATION = os.path.join(os.sep,'var','log','submit','monitors')
MONITORLOGFILENAME = "monitorJob.log"
APPLICATIONLOGGER  = logging.getLogger('')

DBFILENAME   = "monitorJob.db"
DBFILEPATH   = os.path.join(MONITORROOT,'DB',DBFILENAME)
DUMPFILENAME = "monitorJob.sql"
DUMPFILEPATH = os.path.join(MONITORROOT,'DB',DUMPFILENAME)

MONITORSINFOLOCATION = MONITORROOT
MONITORSINFOFILENAME = "monitors.dat"

SSHIDENTITY  = os.path.join(MONITORROOT,".ssh","submit_rsa")
BINDIRECTORY = os.path.join(MONITORROOT,'bin')

ACTIVITYUPDATEINTERVAL = 15.*60.


def openLogger(logDirectory,
               hubLogFile):
   class EmptyFilter(logging.Filter):
      """
      This is a filter which rejects empty messages

      """

      def filter(self,record):
         if record.getMessage() == "":
            emptyRecord = True
         else:
            emptyRecord = False

         return(not emptyRecord)

   APPLICATIONLOGGER.setLevel(logging.DEBUG)

   hubLogPath = os.path.join(logDirectory,hubLogFile)
   logHandler = logging.FileHandler(hubLogPath)
   fdLogFile = logHandler.stream.fileno()

   emptyFilter = EmptyFilter()
   logHandler.addFilter(emptyFilter)

   logFormatter = logging.Formatter('%(asctime)s %(message)s','%s [%a %b %d %H:%M:%S %Y]')
   logHandler.setFormatter(logFormatter)
   APPLICATIONLOGGER.addHandler(logHandler)

   return(fdLogFile)


def daemonize(fdLogFile):
   if fdLogFile != sys.stdout.fileno():
      devnull = open("/dev/null",'r')
      os.dup2(devnull.fileno(),sys.stdin.fileno())
      os.dup2(fdLogFile,sys.stdout.fileno())
      os.dup2(fdLogFile,sys.stderr.fileno())

   if os.fork() != 0:
      os.wait()
      sys.exit(0)
   else:
      os.setsid()
      pid = os.fork()
      if pid != 0:
         sys.exit(0)

   time.sleep(2)


class MonitorJob:
   def __init__(self,
                jobMonitorHost,
                jobMonitorPort,
                tunnelMonitorHost,
                tunnelMonitorPort,
                monitorsInfoLocation,
                monitorsInfoFilename,
                dbFilePath,
                dumpFilePath,
                sshIdentity):
      self.logger               = logging.getLogger(__name__)
      self.jobMonitorHost       = jobMonitorHost
      self.jobMonitorPort       = jobMonitorPort
      self.tunnelMonitorHost    = tunnelMonitorHost
      self.tunnelMonitorPort    = tunnelMonitorPort
      self.monitorsInfoLocation = monitorsInfoLocation
      self.monitorsInfoFilename = monitorsInfoFilename
      self.dbFilePath           = dbFilePath
      self.dumpFilePath         = dumpFilePath
      self.sshIdentity          = sshIdentity

      self.bufferSize        = 4096
      self.siteMonitorPid    = 0
      self.remoteSiteJobInfo = {}
      self.remoteSiteNotify  = {}
      self.inputDescriptors  = []
      self.clientConnections = []

      self.monitorsInfo        = None
      self.jobMonitor          = None
      self.remoteTunnelMonitor = None

      self.reportedJobDoneAgeLimit = 60*60*3
      self.jobDoneAgeLimit         = 60*60*6
      self.purgeJobInterval        = 60*60

      signal.signal(signal.SIGINT,self.sigINT_handler)
      signal.signal(signal.SIGHUP,self.sigHUP_handler)
      signal.signal(signal.SIGQUIT,self.sigQUIT_handler)
      signal.signal(signal.SIGABRT,self.sigABRT_handler)
      signal.signal(signal.SIGTERM,self.sigTERM_handler)


   def terminate(self):
      if self.siteMonitorPid:
         self.logger.log(logging.INFO,getLogMessage("Send TERM to child ssh process"))
         try:
            os.kill(self.siteMonitorPid,signal.SIGTERM)
         except:
            pass
         self.logger.log(logging.INFO,getLogMessage("distributor site monitor stopped"))
      else:
         if self.jobMonitor:
            self.jobMonitor.dumpActiveJobs()
            self.jobMonitor.close()
            self.jobMonitor = None
         for site in self.remoteSiteJobInfo:
            jobsSite,inputFile,outputFile,childPid = self.remoteSiteJobInfo[site]
            self.logger.log(logging.INFO,getLogMessage("Send TERM to child site %s process" % (jobsSite)))
            try:
               os.kill(childPid,signal.SIGTERM)
            except:
               pass
         self.logger.log(logging.INFO,getLogMessage("***********************************"))
         self.logger.log(logging.INFO,getLogMessage("* distributor job monitor stopped *"))
         self.logger.log(logging.INFO,getLogMessage("***********************************"))


   def sigGEN_handler(self,
                      signalNumber,
                      frame):
      self.terminate()
      sys.exit(1)

   
   def sigINT_handler(self,
                      signalNumber,
                      frame):
      self.logger.log(logging.INFO,getLogMessage("Received SIGINT!"))
      del self.monitorsInfo
      self.monitorsInfo = MonitorsInfo(self.monitorsInfoLocation,self.monitorsInfoFilename)
      self.logger.log(logging.INFO,getLogMessage("Site Monitor Info Reloaded!"))
   
   
   def sigHUP_handler(self,
                      signalNumber,
                      frame):
      self.logger.log(logging.INFO,getLogMessage("Received SIGHUP!"))
      self.sigGEN_handler(signalNumber,frame)
   
   
   def sigQUIT_handler(self,
                       signalNumber,
                       frame):
      self.logger.log(logging.INFO,getLogMessage("Received SIGQUIT!"))
      self.sigGEN_handler(signalNumber,frame)
   
   
   def sigABRT_handler(self,
                       signalNumber,
                       frame):
      self.logger.log(logging.INFO,getLogMessage("Received SIGABRT!"))
      self.sigGEN_handler(signalNumber,frame)
   
   
   def sigTERM_handler(self,
                       signalNumber,
                       frame):
      self.logger.log(logging.INFO,getLogMessage("Received SIGTERM!"))
      self.sigGEN_handler(signalNumber,frame)
   

   def getRemoteQueueJobStatus(self,
                               command,
                               inputFd,
                               outputFile):
      commandArgs = shlex.split(command)
      child = subprocess.Popen(commandArgs,bufsize=self.bufferSize,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               close_fds=True)
      self.siteMonitorPid = child.pid
      childin    = child.stdin
      childout   = child.stdout
      childoutFd = childout.fileno()
      childerr   = child.stderr
      childerrFd = childerr.fileno()
   
      outEOF = False
      errEOF = False
   
      outData = []
      errData = []
      dstData = []
   
      while True:
         if os.getppid() == 1:
            try:
               os.kill(os.getpid(),signal.SIGTERM)
            except:
               pass
         toCheck = []
         toCheck.append(inputFd)
         if not outEOF:
            toCheck.append(childoutFd)
         if not errEOF:
            toCheck.append(childerrFd)
         ready = select.select(toCheck,[],[]) # wait for input
   
         if inputFd in ready[0]:
            inChunk = os.read(inputFd,self.bufferSize)
            if inChunk != '':
               childin.write(inChunk)
               childin.flush()
   
         if childoutFd in ready[0]:
            outChunk = os.read(childoutFd,self.bufferSize)
            if outChunk == '':
               outEOF = True
            outData.append(outChunk)
            outputFile.write(outChunk)
            outputFile.flush()
   
         if childerrFd in ready[0]:
            errChunk = os.read(childerrFd,self.bufferSize)
            if errChunk == '':
               errEOF = True
            errData.append(errChunk)
   
         if outEOF and errEOF:
            break
   
      pid,err = os.waitpid(self.siteMonitorPid,0)
      self.siteMonitorPid = 0
      if err != 0:
         if   os.WIFSIGNALED(err):
            self.logger.log(logging.INFO,getLogMessage("%s failed w/ signal %d" % (command,os.WTERMSIG(err))))
         else:
            if os.WIFEXITED(err):
               err = os.WEXITSTATUS(err)
            self.logger.log(logging.INFO,getLogMessage("%s failed w/ exit code %d" % (command,err)))
            self.logger.log(logging.INFO,getLogMessage("%s" % (commandArgs)))
         self.logger.log(logging.INFO,getLogMessage("%s" % ("".join(errData))))

      return(err,"".join(outData),"".join(errData),"\n".join(dstData))
   

   def startNewRemoteMonitor(self,
                             newJobSite):
      sshCommand,tunnelDesignator = self.monitorsInfo.getSSHCommand(newJobSite,self.sshIdentity,self.remoteTunnelMonitor)
      if sshCommand != "":
         parentReceiveFd,childSendFd = os.pipe()
         childReceiveFd,parentSendFd = os.pipe()
   
         pid = os.fork()
         if pid:
   # parent
            try:
               os.close(childReceiveFd)
            except:
               self.logger.log(logging.INFO,getLogMessage("close(childReceiveFd) failed"))
            try:
               os.close(childSendFd)
            except:
               self.logger.log(logging.INFO,getLogMessage("close(childSendFd) failed"))
            inputFile  = os.fdopen(parentReceiveFd,'r')
            outputFile = os.fdopen(parentSendFd,'w')
            self.inputDescriptors.append(parentReceiveFd)
            self.remoteSiteJobInfo[parentReceiveFd] = (newJobSite,inputFile,outputFile,pid)
            self.remoteSiteNotify[newJobSite] = (outputFile)
            self.jobMonitor.addActiveJobSite(newJobSite,time.time())
         else:
   # child
            for channel in self.clientConnections:
               channel.close()
            self.jobMonitor.close()

            try:
               os.close(parentReceiveFd)
            except:
               self.logger.log(logging.INFO,getLogMessage("close(parentReceiveFd) failed"))
            try:
               os.close(parentSendFd)
            except:
               self.logger.log(logging.INFO,getLogMessage("close(parentSendFd) failed"))
            self.logger.log(logging.INFO,getLogMessage("Launching %s" % (newJobSite)))
            outputFile = os.fdopen(childSendFd,'w')
            if tunnelDesignator != "":
               self.remoteTunnelMonitor.incrementTunnelUse(tunnelDesignator)
            self.getRemoteQueueJobStatus(sshCommand,childReceiveFd,outputFile)
            if tunnelDesignator != "":
               self.remoteTunnelMonitor.decrementTunnelUse(tunnelDesignator)
            outputFile.write("0 %s:\n" % (newJobSite))
            outputFile.flush()
            outputFile.close()
            try:
               os.close(childReceiveFd)
            except:
               self.logger.log(logging.INFO,getLogMessage("close(childReceiveFd) failed"))
            self.logger.log(logging.INFO,getLogMessage("Closing %s" % (newJobSite)))
            sys.exit(0)
   

   def monitor(self):
      self.monitorsInfo        = MonitorsInfo(self.monitorsInfoLocation,self.monitorsInfoFilename)
      self.remoteTunnelMonitor = RemoteTunnelMonitor(self.tunnelMonitorHost,self.tunnelMonitorPort)
   
      self.jobMonitor = JobMonitor(self.jobMonitorHost,self.jobMonitorPort,
                                   activeJobDBPath=self.dbFilePath,
                                   activeJobDumpPath=self.dumpFilePath)
      if not self.jobMonitor.isBound():
         self.logger.log(logging.INFO,getLogMessage("Port binding failed"))
         sys.exit(1)

      self.logger.log(logging.INFO,getLogMessage("***********************************"))
      self.logger.log(logging.INFO,getLogMessage("* distributor job monitor started *"))
      self.logger.log(logging.INFO,getLogMessage("***********************************"))
   
      monitorSocketFd = self.jobMonitor.boundFileDescriptor()
      self.inputDescriptors.append(monitorSocketFd)
   
      timeLastPurgeReportedJobDone = 0
      timeLastPurgeJobDone         = 0
      lastReportedActiveJobCount   = 0
   
      while True:
         nPurgeActiveJobs = 0
         now = time.time()
         if now-timeLastPurgeReportedJobDone > self.purgeJobInterval:
            nPurgeActiveJobs += self.jobMonitor.purgeActiveJobs('Dr',self.reportedJobDoneAgeLimit)
            timeLastPurgeReportedJobDone = now
         if now-timeLastPurgeJobDone > self.purgeJobInterval:
            nPurgeActiveJobs += self.jobMonitor.purgeActiveJobs('D',self.jobDoneAgeLimit)
            timeLastPurgeJobDone = now
         if nPurgeActiveJobs > 0:
            self.logger.log(logging.INFO,getLogMessage("%d jobs purged" % (nPurgeActiveJobs)))
            self.jobMonitor.dumpActiveJobs()
         activeJobCount = self.jobMonitor.getActiveJobCount()
         if activeJobCount != lastReportedActiveJobCount:
            self.logger.log(logging.INFO,getLogMessage("%d monitored jobs" % (activeJobCount)))
         lastReportedActiveJobCount = activeJobCount
   
         newJobSite = ""
         newJobId   = ""
         try:
            readyInputFds = select.select(self.inputDescriptors,[],[],ACTIVITYUPDATEINTERVAL)[0]
         except select.error,err:
            if err[0] == EINTR:
               readyInputFds = []
            else:
               for inputDescriptor in self.inputDescriptors:
                  if isinstance(inputDescriptor,socket.socket):
                     try:
                        os.fstat(inputDescriptor.fileno())
                     except:
                        self.logger.log(logging.INFO,getLogMessage(inputDescriptor))
                  else:
                     try:
                        os.fstat(inputDescriptor)
                     except:
                        self.logger.log(logging.INFO,getLogMessage(inputDescriptor))
               self.terminate()
               raise
   
         if monitorSocketFd in readyInputFds:
            channel = self.jobMonitor.acceptConnection()
            if channel:
               self.clientConnections.append(channel)
               self.inputDescriptors.append(channel)
   
         for readyInputFd in readyInputFds:
            if (readyInputFd != monitorSocketFd) and (readyInputFd not in self.clientConnections):
               messageSite,inputFile,outputFile,childPid = self.remoteSiteJobInfo[readyInputFd]
               self.logger.log(logging.INFO,getLogMessage("Update message received from %s" % (messageSite)))
               updateReceived = False
               message = inputFile.readline()
#              log("message: " + message)
   
               messages = message.split(':')
               if messages[0] != "":
                  try:
                     messageLength,messageSite = messages[0].split()
                     messageLength = int(messageLength)
                  except:
                     self.logger.log(logging.ERROR,getLogMessage("Error in message %s" % (messages[0])))
                     messageLength = 0
               else:
                  messageLength = 0
   
               if messageLength == 0:
                  os.waitpid(childPid,0)
                  self.jobMonitor.deleteActiveJobSite(messageSite)
                  self.inputDescriptors.remove(readyInputFd)
                  del self.remoteSiteJobInfo[readyInputFd]
                  del self.remoteSiteNotify[messageSite]
                  inputFile.close()
                  outputFile.close()
                  self.logger.log(logging.INFO,getLogMessage("Closed %s" % (messageSite)))
               else:
                  del messages[0]
                  self.logger.log(logging.INFO,getLogMessage("%d/%d job status messages received from %s" % \
                                                             (len(messages),messageLength,messageSite)))
                  for message in messages:
                     try:
                        remoteJobId,jobStatus,jobStage,jobQueue = (message.split() + ['?'])[:4]
                        jobInfo = JobInfo(messageSite,
                                          remoteJobId,
                                          jobStatus=jobStatus,
                                          jobStage=jobStage,
                                          jobQueue=jobQueue)
                        self.jobMonitor.updateActiveJob(jobInfo)
                     except:
                        pass
                  self.jobMonitor.updateWFJobs(messageSite)
                  self.jobMonitor.dumpActiveJobs()
                  updateReceived = True
               del messages
   
               if updateReceived:
                  self.jobMonitor.updateActiveJobSite(messageSite,time.time())
   
         for channel in readyInputFds:
            if channel in self.clientConnections:
               channelClosed,newJobSite,newJobId = self.jobMonitor.processRequest(channel)
               if channelClosed:
                  self.clientConnections.remove(channel)
                  self.inputDescriptors.remove(channel)
   
               if newJobSite != "":
                  if not self.jobMonitor.isJobSiteActive(newJobSite):
                     self.startNewRemoteMonitor(newJobSite)
   
                  if newJobId != "":
                     try:
                        outputFile = self.remoteSiteNotify[newJobSite]
                        outputFile.write(newJobId + "\n")
                        outputFile.flush()
                     except:
                        pass
   
         self.jobMonitor.updateUserActivityScores()

   
if __name__ == '__main__':

   fdLogFile = openLogger(MONITORLOGLOCATION,MONITORLOGFILENAME)
   daemonize(fdLogFile)

   os.environ['PATH'] = BINDIRECTORY + os.pathsep + os.environ['PATH']

   __monitorJob__ = MonitorJob(JOBMONITORHOST,JOBMONITORPORT,
                               TUNNELMONITORHOST,TUNNELMONITORPORT,
                               MONITORSINFOLOCATION,MONITORSINFOFILENAME,
                               DBFILEPATH,DUMPFILEPATH,
                               SSHIDENTITY)
   __monitorJob__.monitor()


